"""
This module implements the CCR method.
"""

import numpy as np

from sklearn.metrics import pairwise_distances

from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['CCR']

class CCR(OverSampling):
    """
    References:
        * BibTex::

            @article{ccr,
                    author = {Koziarski, Michał and Wozniak, Michal},
                    year = {2017},
                    month = {12},
                    pages = {727–736},
                    title = {CCR: A combined cleaning and resampling algorithm
                                for imbalanced data classification},
                    volume = {27},
                    journal = {International Journal of Applied Mathematics
                                and Computer Science}
                    }

    Notes:
        * Adapted from https://github.com/michalkoziarski/CCR
        * The number of synthetic samples to generate is hacked for reasonable results
    """

    categories = [OverSampling.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 *,
                 energy=1.0,
                 scaling=0.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            energy (float): energy parameter
            scaling (float): scaling factor
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(energy, "energy", 0)
        self.check_greater_or_equal(scaling, "scaling", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.energy = energy
        self.scaling = scaling
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'energy': [0.001, 0.0025, 0.005,
                                             0.01, 0.025, 0.05, 0.1,
                                             0.25, 0.5, 1.0, 2.5, 5.0,
                                             10.0, 25.0, 50.0, 100.0],
                                  'scaling': [0.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def taxicab_sample(self, n_dim, radius):
        """
        Generate one taxi cab sample.

        Args:
            n_dim (int): the number of dimensions
            radius (float): the radius to be used

        Returns:
            np.array: one taxicab sample
        """
        sample = []
        random_numbers = self.random_state.rand(n_dim)

        for idx in range(n_dim):
            # spread = r - np.sum(np.abs(sample))
            spread = radius
            if len(sample) > 0:
                spread -= abs(sample[-1])
            sample.append(spread * (2 * random_numbers[idx] - 1))

        return self.random_state.permutation(sample)

    def taxicab_samples(self, n_dim, radius, n_samples=1):
        """
        Generate n taxicab samples.

        Args:
            n_dim (int): the number of dimensions.
            radius (float): the radius to be used
            n_samples (int): the number of samples to be generated

        Returns:
            np.array: the generated samples
        """
        return np.vstack([self.taxicab_sample(n_dim, radius) \
                                        for _ in range(n_samples)])

    def random_majority_offset(self, n_dim):
        """
        Generate random majority offset.

        Args:
            n_dim (int): number of dimensions

        Returns:
            np.array: the random majority offset
        """
        r_num = self.random_state.rand(n_dim)
        r_num = 1e-6 * r_num + 1e-6
        r_sign = self.random_state.choice([-1.0, 1.0], n_dim)

        return r_num * r_sign

    def update_majority_point(self, dist, majority_point, minority_point):
        """
        Updating the majority point to break up ties with the minority point

        Args:
            float (dist): the distance
            majority_point (np.array): a majority point
            minority_pont (np.array): a minority point

        Returns:
            float, np.array: the updated dist and majority_point
        """
        if dist < 1e-20:
            majority_point = majority_point + \
                            self.random_majority_offset(majority_point.shape[0])
            dist = np.sum(np.abs(minority_point - majority_point))
            return dist, majority_point
        return dist, majority_point

    def determine_radii_current_majority(self, majority, distances, idx, sorted_distances):
        """
        Determines the radius for the minority sample idx and also returns the
        majority (current_majority) the algorithm arrives to.

        Args:
            majority (np.array): majority samples
            distances (np.array): the distance matrix
            idx (int): the index of the minority samples to work on
            sorted_distances (np.array): the sorted distances of the minority
                                            item
        """
        last_distance = 0.0
        remaining_energy = self.energy * (majority.shape[1] ** self.scaling)
        radius = 0.0
        current_majority = 0
        for current_majority in range(len(majority)):
            if remaining_energy <= 0:
                remaining_energy = 0
                break

            radius_change = remaining_energy / (current_majority + 1.0)

            dist = distances[idx, sorted_distances[current_majority]]
            if dist >= radius + radius_change:
                radius += radius_change
                break

            curr_maj_idx = sorted_distances[current_majority]
            radius_change = np.max([distances[idx, curr_maj_idx] - last_distance, 0])
            radius += radius_change
            remaining_energy -= radius_change * (current_majority + 1.0)

            last_distance = distances[idx, sorted_distances[current_majority - 1]]


        # if current_majority == len(majority) - 1 then adjust
        multiplier_flag = int(current_majority == (len(majority) - 1))
        radius += (remaining_energy / max(1, current_majority)) * multiplier_flag

        return radius, current_majority

    def determine_radii_translations(self, minority, majority):
        """
        Determine the radii and the translations.

        Args:
            minority (np.array): the minority samples
            majority (np.array): the majority samples

        Returns:
            np.array, np.array: the radii and the translations
        """
        distances = pairwise_distances(minority, majority, metric='l1')

        radii = np.zeros(len(minority))
        translations = np.zeros(majority.shape)

        for idx, minority_point in enumerate(minority):
            sorted_distances = np.argsort(distances[idx])
            radius, current_majority = self.determine_radii_current_majority(majority,
                                                                            distances,
                                                                            idx,
                                                                            sorted_distances)

            radii[idx] = radius

            for jdx in range(current_majority):
                majority_point = majority[sorted_distances[jdx]]
                dist = distances[idx, sorted_distances[jdx]]

                dist, majority_point = self.update_majority_point(dist,
                                                                majority_point,
                                                                minority_point)

                translation = (radius - dist) / dist * (majority_point - minority_point)
                translations[sorted_distances[jdx]] += translation

        return radii, translations

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        minority = X[y == self.min_label]
        majority = X[y == self.maj_label]

        radii, translations = self.determine_radii_translations(minority,
                                                                majority)

        majority = majority.astype(float)
        majority += translations

        # this is changed to the ratio of sums for more robustness
        radii_inv_sum = len(radii) / np.sum(radii)
        #radii_inv_sum = 1.0 / np.sum(1.0 / radii)

        appended = [np.zeros(shape=(0, majority.shape[1]))]
        for idx, minority_point in enumerate(minority):
            # this determination of synthetic samples is hacked
            # to give reasonable results
            synthetic_samples = n_to_sample / len(minority) * (1.0 / radii[idx]) / radii_inv_sum
            synthetic_samples = int(np.round(synthetic_samples))
            if synthetic_samples > 0:
                tc_samples = self.taxicab_samples(minority_point.shape[0],
                                                    radii[idx],
                                                    synthetic_samples)
                appended.append(minority_point + tc_samples)

        appended = np.vstack(appended)

        return (np.vstack([X, appended]),
                np.hstack([y, np.repeat(self.min_label, len(appended))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'energy': self.energy,
                'scaling': self.scaling,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the reproducibility of oversampling and noise filtering.
"""

import numpy as np
import pytest

import smote_variants as sv

from smote_variants.datasets import load_normal
from .additional_objs import additional_objs

dataset = load_normal()

X_normal = dataset['data']
y_normal = dataset['target']

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_reproducibility(smote_class):
    """
    Tests the reproducibility of oversampling.

    Args:
        smote_class (class): an oversampling class
    """
    X_orig = X_normal.copy()
    y_orig = y_normal.copy()

    X_a, y_a = smote_class(random_state=5).sample(X_normal, y_normal)
    oversampler = smote_class(random_state=5)
    X_b, y_b = oversampler.sample(X_normal, y_normal)
    X_c, y_c = smote_class(**oversampler.get_params()).sample(X_normal, y_normal)

    assert np.array_equal(X_a, X_b)
    assert np.array_equal(X_b, X_c)
    assert np.array_equal(X_orig, X_normal)

    assert np.array_equal(y_a, y_b)
    assert np.array_equal(y_b, y_c)
    assert np.array_equal(y_orig, y_normal)

@pytest.mark.parametrize("smote_obj", additional_objs)
def test_reproducibility_additional(smote_obj):
    """
    Tests the reproducibility of specific parameterizations.

    Args:
        smote_obj (obj): an oversampling object
    """
    smote_obj = smote_obj.__class__(**smote_obj.get_params())

    X_orig = X_normal.copy()
    y_orig = y_normal.copy()

    X_a, y_a = smote_obj.sample(X_normal, y_normal)
    X_c, y_c = smote_obj.__class__(**smote_obj.get_params()).sample(X_normal, y_normal)

    assert np.array_equal(X_a, X_c)
    assert np.array_equal(y_a, y_c)
    assert np.array_equal(X_normal, X_orig)
    assert np.array_equal(y_normal, y_orig)

@pytest.mark.parametrize("nf_class", sv.get_all_noisefilters())
def test_reproducibility_noisefilters(nf_class):
    """
    Tests the reproducibility of noise filtering.

    Args:
        nf_class (class): a noise filtering class
    """
    X_orig = X_normal.copy()
    y_orig = y_normal.copy()

    X_a, y_a = nf_class().remove_noise(X_normal, y_normal)
    noisefilter = nf_class()
    X_b, y_b = noisefilter.remove_noise(X_normal, y_normal)
    X_c, y_c = nf_class(**noisefilter.get_params()).remove_noise(X_normal, y_normal)

    assert np.array_equal(X_a, X_b)
    assert np.array_equal(X_b, X_c)
    assert np.array_equal(X_normal, X_orig)

    assert np.array_equal(y_a, y_b)
    assert np.array_equal(y_b, y_c)
    assert np.array_equal(y_normal, y_orig)



"""
Testing the evaluation.
"""

import os
import shutil

from sklearn import datasets

from smote_variants.evaluation import Folding, SamplingJob, EvaluationJob
import smote_variants

DATASET = datasets.load_breast_cancer()
DATASET['name'] = 'breast_cancer'

OVERSAMPLER = 'SMOTE'
OVERSAMPLER_PARAMS = {}

CLASSIFIERS = [('sklearn.tree', 'DecisionTreeClassifier', {})]

CACHE_PATH = os.path.join('.', 'test_path')

def test_evaluation():
    """
    Testing the evaluation.
    """
    shutil.rmtree(CACHE_PATH, ignore_errors=True)

    folding = Folding(DATASET,
                    cache_path=CACHE_PATH,
                    reset=True)

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        OVERSAMPLER,
                        OVERSAMPLER_PARAMS,
                        cache_path=CACHE_PATH)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            CLASSIFIERS,
                            cache_path=CACHE_PATH)

        result = ejob.do_evaluation()

        assert os.path.exists(result[0])

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        OVERSAMPLER,
                        OVERSAMPLER_PARAMS,
                        cache_path=CACHE_PATH)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            CLASSIFIERS,
                            cache_path=CACHE_PATH)

        result = ejob.do_evaluation()

        assert os.path.exists(result[0])

    shutil.rmtree(CACHE_PATH)

def test_evaluation_warning():
    """
    Testing the evaluation warning.
    """
    folding = Folding(DATASET,
                    reset=True)

    smote_variants.config.suppress_internal_warnings(False)

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        OVERSAMPLER,
                        OVERSAMPLER_PARAMS)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            [('smote_variants.classifiers',
                             'ErrorWarningClassifier',
                             {'raise_warning': True})])

        result = ejob.do_evaluation()

        assert len(result[0]['warning']) > 3

        smote_variants.config.suppress_internal_warnings(True)

def test_evaluation_error():
    """
    Testing the evaluation error.
    """
    folding = Folding(DATASET,
                    reset=True)

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        OVERSAMPLER,
                        OVERSAMPLER_PARAMS)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            [('smote_variants.classifiers',
                             'ErrorWarningClassifier',
                             {'raise_value_error': True})])

        result_eval = ejob.do_evaluation()

        assert len(result_eval[0]['error']) > 3

        ejob = EvaluationJob(result,
                            [('smote_variants.classifiers',
                             'ErrorWarningClassifier',
                             {'raise_runtime_error': True})])

        result_eval = ejob.do_evaluation()

        assert len(result_eval[0]['error']) > 3

def test_sampling_cached_folding():
    """
    Testing the sampling of a cached folding.
    """
    shutil.rmtree(CACHE_PATH, ignore_errors=True)

    folding = Folding(DATASET,
                    cache_path=CACHE_PATH,
                    reset=True)
    oversampler = 'SMOTE'
    oversampler_params = {}

    for fold in folding.folding_files():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params,
                        cache_path=CACHE_PATH)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            CLASSIFIERS,
                            cache_path=CACHE_PATH)

        result = ejob.do_evaluation()

        assert os.path.exists(result[0])

    for fold in folding.folding_files():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params,
                        cache_path=CACHE_PATH)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            CLASSIFIERS,
                            cache_path=CACHE_PATH)

        result = ejob.do_evaluation()

        assert os.path.exists(result[0])

    shutil.rmtree(CACHE_PATH)

def test_sampling_no_caching():
    """
    Testing sampling with no caching.
    """
    folding = Folding(DATASET,
                    reset=True)
    oversampler = 'SMOTE'
    oversampler_params = {}

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params)

        result = sjob.do_oversampling()

        ejob = EvaluationJob(result,
                            CLASSIFIERS)

        result = ejob.do_evaluation()

        assert isinstance(result[0], dict)

def test_get_params():
    """
    Test the get_params function.
    """
    ejob = EvaluationJob({},
                        CLASSIFIERS,
                        cache_path=CACHE_PATH)

    assert len(ejob.get_params()) > 0



"""
This module implements the DE_oversampling method.
"""
import warnings

import numpy as np

from sklearn.cluster import KMeans
from ..config import suppress_external_warnings
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['DE_oversampling']

class DE_oversampling(OverSampling): # pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{de_oversampling,
                            author={Chen, L. and Cai, Z. and Chen, L. and
                                    Gu, Q.},
                            booktitle={2010 Third International Conference
                                       on Knowledge Discovery and Data Mining},
                            title={A Novel Differential Evolution-Clustering
                                    Hybrid Resampling Algorithm on Imbalanced
                                    Datasets},
                            year={2010},
                            volume={},
                            number={},
                            pages={81-85},
                            keywords={pattern clustering;sampling methods;
                                        support vector machines;differential
                                        evolution;clustering algorithm;hybrid
                                        resampling algorithm;imbalanced
                                        datasets;support vector machine;
                                        minority class;mutation operators;
                                        crossover operators;data cleaning
                                        method;F-measure criterion;ROC area
                                        criterion;Support vector machines;
                                        Intrusion detection;Support vector
                                        machine classification;Cleaning;
                                        Electronic mail;Clustering algorithms;
                                        Signal to noise ratio;Learning
                                        systems;Data mining;Geology;imbalanced
                                        datasets;hybrid resampling;clustering;
                                        differential evolution;support vector
                                        machine},
                            doi={10.1109/WKDD.2010.48},
                            ISSN={},
                            month={Jan},}
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_uses_clustering,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params={},
                 crossover_rate=0.5,
                 similarity_threshold=0.5,
                 n_clusters=30,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): control parameter of the nearest neighbor
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            crossover_rate (float): cross over rate of evoluation
            similarity_threshold (float): similarity threshold parameter
            n_clusters (int): number of clusters for cleansing
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 2)
        self.check_in_range(crossover_rate, 'crossover_rate', [0, 1])
        self.check_in_range(similarity_threshold,
                            'similarity_threshold', [0, 1])
        self.check_greater_or_equal(n_clusters, 'n_clusters', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.crossover_rate = crossover_rate
        self.similarity_threshold = similarity_threshold
        self.n_clusters = n_clusters
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'crossover_rate': [0.1, 0.5, 0.9],
                                  'similarity_threshold': [0.5, 0.9],
                                  'n_clusters': [10, 20, 50]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self, n_to_sample, X_min, indices):
        """
        Generate samples

        Args:
            n_to_sample (int): number of samples to generate
            X_min (np.array): minority samples
            indices (np.array): the neighborhood graph

        Returns:
            np.array: the generated samples
        """
        n_dim = X_min.shape[1]
        samples = []
        for _ in range(n_to_sample):
            # mutation according to the description in the paper
            random_index = self.random_state.randint(len(X_min))
            random_point = X_min[random_index]
            random_neighbor_indices = self.random_state.choice(
                indices[random_index][1:], 2, replace=False)
            random_neighbor_1 = X_min[random_neighbor_indices[0]]
            random_neighbor_2 = X_min[random_neighbor_indices[1]]

            mutated = random_point + \
                (random_neighbor_1 - random_neighbor_2) * \
                self.random_state.random_sample()

            # crossover - updates the vector 'mutated'
            rand_s = self.random_state.randint(n_dim)
            for idx in range(n_dim):
                random_value = self.random_state.random_sample()
                if random_value >= self.crossover_rate and not idx == rand_s:
                    mutated[idx] = random_point[idx]
                elif random_value < self.crossover_rate or idx == rand_s:
                    pass

            samples.append(mutated)
        return samples

    def cleansing(self, X, y):
        """
        Cleansing

        Args:
            X (np.array): the extended features
            y (np.array): the extended target labels

        Returns:
            np.array: indices to remove
        """
        # cleansing based on clustering
        kmeans = KMeans(n_clusters=min([len(X), self.n_clusters]),
                        random_state=self._random_state_init)

        with warnings.catch_warnings():
            if suppress_external_warnings():
                warnings.simplefilter("ignore")
            kmeans.fit(X)

        def cluster_filter(label):
            return len(np.unique(y[np.where(kmeans.labels_ == label)[0]])) == 1

        one_label_clusters = [label \
                    for label in np.unique(kmeans.labels_) if cluster_filter(label)]
        to_remove = []

        # going through the clusters having one label only
        for label in one_label_clusters:
            cluster_indices = np.where(kmeans.labels_ == label)[0]
            mean_of_cluster = kmeans.cluster_centers_[label]

            # finding center-like sample
            tmp = np.linalg.norm(X[cluster_indices] - mean_of_cluster, axis=1)
            center_like_index = np.argmin(tmp)
            center_like_index = cluster_indices[center_like_index]

            # removing the samples similar to the center-like sample
            for idx in cluster_indices:
                if idx != center_like_index:
                    denom = (np.linalg.norm(X[idx]) *
                                        np.linalg.norm(X[center_like_index]))
                    num = np.inner(X[idx], X[center_like_index])

                    if denom == 0.0 or (num / denom) > self.similarity_threshold:
                        to_remove.append(idx)

        return to_remove

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        n_neighbors = min([len(X_min), self.n_neighbors+1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        indices = nnmt.kneighbors(X_min, return_distance=False)

        # generating samples
        samples = self.generate_samples(n_to_sample, X_min, indices)

        # assembling all data for clearning
        X, y = np.vstack([X, np.vstack(samples)]), np.hstack(
            [y, np.repeat(self.min_label, len(samples))])

        to_remove = self.cleansing(X, y)

        return np.delete(X, to_remove, axis=0), np.delete(y, to_remove)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'crossover_rate': self.crossover_rate,
                'similarity_threshold': self.similarity_threshold,
                'n_clusters': self.n_clusters,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the oversampling classifier.
"""

from sklearn import datasets

from smote_variants.base import equal_dicts
from smote_variants.classifiers import OversamplingClassifier, ErrorWarningClassifier

dataset = datasets.load_breast_cancer()

def test_oversamplingclassifier():
    """
    Testing the oversampling classifier
    """
    osc = OversamplingClassifier(oversampler=('smote_variants', 'SMOTE', {}),
                                    classifier=('sklearn.neighbors', 'KNeighborsClassifier', {}))

    osc.fit(dataset['data'], dataset['target'])

    pred = osc.predict(dataset['data'])

    assert len(pred) == len(dataset['data'])

    prob = osc.predict_proba(dataset['data'])

    assert len(prob) == len(dataset['data'])

    params = osc.get_params()

    assert len(params) > 0

    osc.set_params(**params)

    assert equal_dicts(params, osc.get_params())

    assert len(ErrorWarningClassifier().get_params()) > 0



"""
This module implements the SMOTE_FRST_2T method.
"""

import numpy as np

from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import pairwise_distances

from ..base import coalesce, coalesce_dict
from ..base import OverSampling

from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['SMOTE_FRST_2T']

class SMOTE_FRST_2T(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_frst_2t,
                        title = "Fuzzy-rough imbalanced learning for the
                                    diagnosis of High Voltage Circuit
                                    Breaker maintenance: The SMOTE-FRST-2T
                                    algorithm",
                        journal = "Engineering Applications of Artificial
                        Intelligence",
                        volume = "48",
                        pages = "134 - 139",
                        year = "2016",
                        issn = "0952-1976",
                        doi = "https://doi.org/10.1016/j.engappai.2015.10.009",
                        author = "Ramentol, E. and Gondres, I. and Lajes, S.
                                    and Bello, R. and Caballero,Y. and
                                    Cornelis, C. and Herrera, F.",
                        keywords = "High Voltage Circuit Breaker (HVCB),
                                    Imbalanced learning, Fuzzy rough set
                                    theory, Resampling methods"
                        }

    Notes:
        * Unlucky setting of parameters might result 0 points added, we have
            fixed this by increasing the gamma_S threshold if the number of
            samples accepted is low.
        * Similarly, unlucky setting of parameters might result all majority
            samples turned into minority.
        * In my opinion, in the algorithm presented in the paper the
            relations are incorrect. The authors talk about accepting samples
            having POS score below a threshold, and in the algorithm in
            both places POS >= gamma is used.
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_sample_ordinary,
                  OverSampling.cat_application,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 gamma_S=0.7, # pylint: disable=invalid-name
                 gamma_M=0.03, # pylint: disable=invalid-name
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            n_neighbors (int): number of neighbors in the SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            gamma_S (float): threshold of synthesized samples
            gamma_M (float): threshold of majority samples
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(gamma_S, "gamma_S", 0)
        self.check_greater_or_equal(gamma_M, "gamma_M", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.gamma_S = gamma_S # pylint: disable=invalid-name
        self.gamma_M = gamma_M # pylint: disable=invalid-name
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'n_neighbors': [3, 5, 7],
                                  'gamma_S': [0.8, 1.0],
                                  'gamma_M': [0.03, 0.05, 0.1]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def pos_cache(self,
                    X_0, # pylint: disable=invalid-name
                    X_1 # pylint: disable=invalid-name
                    ):
        """
        Calculate the matrix used to derive the positive membership values

        Args:
            X_0 (np.array): one array of vectors
            X_1 (np.array): another array of vectors

        Returns:
            np.array: the matrix that can be used to calculate the
                        membership values of X_1
        """
        # because of the minmax scaling, the clipping in eq. 1 is not needed
        dist = pairwise_distances(X_0, X_1, metric='l1')
        R_xy = X_0.shape[1] - dist # pylint: disable=invalid-name
        # taking the Lukasiewitcz T-norm
        R_xy = R_xy - X_0.shape[1] + 1.0 # pylint: disable=invalid-name
        R_xy[R_xy < 0.0] = 0.0

        pos = 1.0 - R_xy

        return pos

    def iteration(self, X, y, X_min, smote):
        """
        The function implementing the main iteration of the algorithm

        Args:
            X (np.array): the training vectors
            y (np.array): the target labels
            X_min (np.array): the minority samples
            smote (obj): the smote object used for sampling

        Returns:
            np.array, np.array: the updated majority and new minority samples
        """
        X_maj = X[y == self.maj_label]

        pos_cache = self.pos_cache(X_min, X_maj)

        # initializing some lists containing the results
        result_synth = np.zeros(shape=(0, X.shape[1]))
        result_maj = np.zeros(shape=(0, X.shape[1]))
        iteration = 0

        gammas = (self.gamma_S, self.gamma_M)

        while ((len(X_min) + len(result_synth) + len(result_maj)) < len(X_maj)
                    and iteration < 100):
            _logger.info("%s: iteration: %d",
                            self.__class__.__name__, iteration)

            # checking if the parameters aren't too conservative
            if len(result_synth) < iteration:
                gammas = (gammas[0] * 1.1, gammas[1])
                _logger.info("%s: gamma_S increased to %f",
                            self.__class__.__name__, gammas[0])

            X_samp = smote.sample(X, y)[0][X.shape[0]:]

            # computing POS membership values for the new samples
            pos_synth = self.pos_cache(X_min, X_samp)

            # adding samples with POS membership smaller than gamma_S to the
            # minority set
            to_add = np.where(np.min(pos_synth, axis=0) < gammas[0])[0]
            result_synth = np.vstack([result_synth, X_samp[to_add]])

            # checking the minimum POS values of the majority samples
            to_remove = np.where(np.min(pos_cache, axis=0) < gammas[1])[0]

            # if the number of majority samples with POS membership smaller
            # than gamma_M is not extreme, then changing labels, otherwise
            # decreasing gamma_M
            if len(to_remove) > (len(X_maj) - len(X_min))/2:
                to_remove = np.array([])
                gammas = (gammas[0], gammas[1] * 0.9)
                _logger.info("%s: gamma_M decreased to %f",
                                self.__class__.__name__, gammas[1])
            elif len(to_remove) > 0:
                result_maj = np.vstack([result_maj, X_maj[to_remove]])
                X_maj = np.delete(X_maj, to_remove, axis=0)
                pos_cache = np.delete(pos_cache, to_remove, axis=1)
                pos_cache = np.vstack([pos_cache,
                                self.pos_cache(result_maj[-len(to_remove):],
                                                X_maj)])

            # updating pos cache
            if len(to_add) > 0:
                pos_cache = np.vstack([pos_cache,
                                self.pos_cache(X_samp[to_add],
                                                X_maj)])

            _logger.info("%s: minority added: %d, majority removed: %d",
                        self.__class__.__name__, len(to_add), len(to_remove))
            iteration = iteration + 1

        return X_maj, np.vstack([result_synth, result_maj])

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        if np.sum(y == self.maj_label) <= np.sum(y == self.min_label):
            return self.return_copies(X, y, "Sampling is not needed")

        # Turning the ranges to 1 speeds up the positive membership
        # calculations
        mmscaler = MinMaxScaler()
        X = mmscaler.fit_transform(X)

        X_min = X[y == self.min_label]

        # after MinMax scaling, the POS value can be calculated as follows
        # determine proportion
        diff = np.sum(y == self.maj_label) - np.sum(y == self.min_label)
        prop = np.max([1.1 / diff, 0.2])

        # executing SMOTE to generate some minority samples
        smote = SMOTE(proportion=prop,
                        n_neighbors=self.n_neighbors,
                        nn_params=self.nn_params,
                        ss_params=self.ss_params,
                        n_jobs=self.n_jobs,
                        random_state=self._random_state_init)

        # iterating until the dataset becomes balanced
        X_maj, result_min = self.iteration(X, y, X_min, smote)

        # packing the results
        X_res = np.vstack([X_maj, X_min, result_min]) # pylint: disable=invalid-name

        y_res_maj = np.repeat(self.maj_label, len(X_maj))
        y_res_min = np.repeat(self.min_label, len(X_res) - len(X_maj))
        y_res = np.hstack([y_res_maj, y_res_min])

        return mmscaler.inverse_transform(X_res), y_res

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'gamma_S': self.gamma_S,
                'gamma_M': self.gamma_M,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the parameterization.
"""

import numpy as np
import pytest

import smote_variants as sv

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_parameters(smote_class):
    """
    Test the parameterization.

    Args:
        class_name (class): an oversampling class
    """
    random_state = np.random.RandomState(5)

    par_comb = smote_class.parameter_combinations()

    original_parameters = random_state.choice(par_comb)
    oversampler = smote_class(**original_parameters)
    parameters = oversampler.get_params()

    assert all(v == parameters[k] for k, v in original_parameters.items())



"""
This module contains all experimental and stable functionalities related
to metric learning.
"""
import warnings

import numpy as np

from sklearn.neighbors import NearestNeighbors
from sklearn.linear_model import LinearRegression
from sklearn.feature_selection import mutual_info_regression

from scipy.stats import rankdata

from metric_learn import ITML_Supervised, LSML_Supervised, NCA

from ._base import instantiate_obj, coalesce
from .._logger import logger
_logger = logger

__all__= ['NearestNeighborsWithMetricTensor',
          'ClassifierImpliedDissimilarityMatrix',
          'MetricTensor',
          'MetricLearningMixin',
          'pairwise_distances_mahalanobis',
          'distances_mahalanobis',
          'fix_pd_matrix',
          'reverse_matrix',
          'construct_tensor',
          'ClosestNeighborsInClasses',
          'RemoveCorrelatedColumns',
          'discrete_variable_mask',
          'estimate_mutual_information',
          'n_neighbors_func',
          'psd_mean']

def pairwise_distances_mahalanobis(X, *, Y=None, tensor=None):
    """
    Returns the pairwise distances of two arrays of vectors using the
    provided metric tensor matrix. If no Y paramter is provided, the
    distance matrix of the paramter X is returned.

    Args:
        X (np.array): array of vectors to calculate distances for
        Y (np.array/None): optional second array of vectors
        tensor (np.array): the metric tensor to be used

    Returns:
        np.array: the distance matrix
    """
    if Y is None:
        Y = X
    if tensor is None:
        tensor = np.eye(X.shape[1])
    tmp= (X[:,None] - Y)
    return np.sqrt(np.einsum('ijk,ijk -> ij', tmp, np.dot(tmp, tensor)))

def distances_mahalanobis(X, Y, tensor=None):
    """
    Returns the distances of the corresponding vectors.

    Args:
        X (np.array): array of vectors
        Y (np.array): another array of vectors
        tensor (np.array): the metric tensor to be used

    Returns:
        np.array: the distances
    """
    if tensor is None:
        tensor = np.eye(X.shape[1])
    tmp = X - Y
    return np.sqrt(np.einsum('ij,jk,ik -> i', tmp, tensor, tmp))

def fix_pd_matrix(matrix, eps=1e-4):
    """
    Fix a positive definite matrix

    Args:
        matrix (np.array): the matrix to be fixed
        eps (float): the tolerance

    Returns:
        np.array: the fixed matrix
    """
    eigv, eigw= np.linalg.eigh(matrix)
    eigv = np.real(eigv)
    eigw = np.real(eigw)
    eigv[eigv <= eps]= eps
    matrix= np.dot(np.dot(eigw, np.diag(eigv)), eigw.T)
    return matrix

def reverse_matrix(matrix):
    """
    Reverse the order of eigenvalues.

    Args:
        matrix (np.array): the matrix to transform

    Returns:
        np.array: the transformed matrix
    """
    eigv, eigw = np.linalg.eigh(matrix)
    eigv = eigv[np.argsort(eigv)[len(eigv) - rankdata(eigv).astype(int)]]
    matrix = np.dot(np.dot(eigw, np.diag(eigv)), eigw.T)
    return matrix

def create_metric_tensor(n_dims, tu_indices, elements):
    """
    Create metric tensor from upper triangle matrix elements.

    Args:
         n_dims (int): dimensionality
         tu_indices (np.array): indices of the upper triangle
         elements (np.array): the values corresponding to the indices

    Returns:
        np.array: the constructed distance matrix
    """
    # creating the metric tensor
    dist= np.zeros((n_dims, n_dims))
    # injecting the regressed coefficients into the upper triangle
    dist[tu_indices]= elements
    # copying by transposing
    dist= (dist + dist.T)
    # halving the elements in the main diagonal due to duplication
    dist[np.diag_indices(n_dims)]= np.diag(dist)/2
    # creating a valid positive definite matrix
    dist= fix_pd_matrix(dist)

    return dist


def construct_tensor(X, dissim_matrix):
    """
    Construct tensor from dissimilarity matrix.

    Args:
        X (np.array): the data to use
        dissim_matrix (np.array): the dissimilarity matrix

    Returns:
        np.array: the metric tensor
    """

    # pre-calculating some triangle indices
    X_tu_indices= np.triu_indices(X.shape[0], k=0)
    d_tu_indices_0= np.triu_indices(X.shape[1], k=0)
    d_tu_indices_1= np.triu_indices(X.shape[1], k=1)

    n_upper, n_d= len(X_tu_indices[0]), len(d_tu_indices_0[0])

    # calculating the cross differences and extracting the upper triangle into
    # a row-major representation
    cross_diff_all= (X[:,None] - X)[X_tu_indices]

    # if the total number of pairs with distances is much greater than the
    # number of free components of the metric tensor, the samples prepared for
    # regression are sampled for efficiency
    mask= np.repeat(True, len(cross_diff_all))
    if n_upper > n_d*100:
        rng = np.arange(n_upper)
        mask[np.random.RandomState(5).permutation(rng)[n_d*100:]]= False

    # preparing the dissimilarity values for regression
    y_target= dissim_matrix[(X_tu_indices[0][mask], X_tu_indices[1][mask])]**2

    # calculating the cross product of components for each pair in the cross
    # difference matrix
    cross_diff_all= cross_diff_all[mask]
    cross_diff_cross_products= np.einsum('...i,...j->...ij',
                                        cross_diff_all,
                                        cross_diff_all)

    # adjustment due to extracting the upper triangle only
    cross_diff_cross_products[:, d_tu_indices_1[0], d_tu_indices_1[1]]*= 2

    # preparing the components of the cross products of pairwise distances
    # for regression
    X_target= cross_diff_cross_products[:, d_tu_indices_0[0], d_tu_indices_0[1]]

    # calculating the elements of the metric tensor by regression
    linearr= LinearRegression(fit_intercept=False).fit(X_target, y_target)

    # creating the metric tensor
    metric_tensor= create_metric_tensor(X.shape[1],
                                        d_tu_indices_0,
                                        linearr.coef_)

    return metric_tensor, linearr.score(X_target, y_target)


class ClassifierImpliedDissimilarityMatrix:
    """
    Computes classifier implied dissimilarity matrix
    """
    def __init__(self,
                 classifier=('sklearn.ensemble',
                            'RandomForestClassifier'),
                 classifier_params=None):
        """
        Constructor of the object

        Args:
            classifier (str): name of a classifier class (available in sklearn)
            classifier_params (dict): parameters of the classifier
        """
        self.classifier= classifier
        if classifier_params is None:
            classifier_params = {'n_estimators': 100,
                                    'min_samples_leaf': 2,
                                    'random_state': 5}
        self.classifier_params= classifier_params
        self.classifier_obj = None
        self.terminal_nodes = None

    def get_params(self):
        """
        The parameters of the object.

        Returns:
            dict: the parameters of the object
        """
        return {'classifier': self.classifier,
                'classifier_params': self.classifier_params}

    def fit(self, X, y):
        """
        Fit the object.

        Args:
            X (np.array): the X vectors
            y (np.array): the target vectors

        Returns:
            ClassifierImpliedDissimilarityMatrix: the fitted object
        """
        _logger.info("%s: fitting", self.__class__.__name__)
        self.classifier_obj = instantiate_obj((self.classifier[0],
                                                self.classifier[1],
                                                self.classifier_params))
        self.classifier_obj.fit(X, y)
        return self

    def transform_kneighbors(self, X):
        """
        Transform using kneighbors.

        Args:
            X (np.array): the vectors to transform

        Returns:
            np.array: the transformed vectors
        """
        _logger.info("%s: transform kneighbors", self.__class__.__name__)
        tnodes= self.classifier_obj.apply(X)
        tmp_nodes= np.vstack([self.terminal_nodes, tnodes])

        def func(vector):
            return 1*np.equal.outer(vector,
                                    vector)[-len(tnodes):][:, :-len(tnodes)]

        results= 1.0 - np.apply_along_axis(func, 0, tmp_nodes)
        results = results.sum(axis=2)/tmp_nodes.shape[1]

        return results

    def dissimilarity_matrix(self, X):
        """
        Calculates the dissimilarity matrix, if y is None, then uses the
        already fitted model. First needs to be called with valid y vector.

        Args:
            X (np.ndarray): explanatory variables
            y (np.array): target (class labels)
        """
        # terminal nodes: rows - samples, columns - trees in the forest
        self.terminal_nodes= self.classifier_obj.apply(X)

        def func(vector):
            return 1*np.equal.outer(vector, vector)

        result =  1.0 - np.apply_along_axis(func,
                                            0,
                                            self.terminal_nodes)
        return result.sum(axis=2)/self.terminal_nodes.shape[1]


class ClosestNeighborsInClasses:
    """
    Closest neighbors within class.
    """
    def __init__(self, n_neighbors=5):
        """
        Constructor of the class, determines the border points.

        Args:
            n_neighbors (int): number of neighbors
        """
        self.n_neighbors= n_neighbors

    def get_params(self):
        """
        Get the dict of parameters

        Returns:
            dict: the dict of parameters
        """
        return {'n_neighbors': self.n_neighbors}

    def fit_transform(self, X, y):
        """
        Fit and transform the dataset

        Args:
            X (np.array): the vectors to operate on
            y (np.array): the corresponding labels

        Returns:
            np.array, np.array: the border points
        """
        X_min= X[y == 1]
        X_maj= X[y == 0]

        nearestn= NearestNeighbors(n_neighbors=self.n_neighbors).fit(X)
        _, ind_min= nearestn.kneighbors(X_min)
        _, ind_maj= nearestn.kneighbors(X_maj)

        label_min= np.all((y[ind_min] == 1), axis=1)
        label_maj= np.all((y[ind_maj] == 0), axis=1)

        X_final = np.vstack([X_maj[~label_maj], X_min[~label_min]])
        y_final = np.hstack([np.repeat(0, int(np.sum(label_maj))),
                             np.repeat(1, int(np.sum(label_min)))])

        return X_final, y_final


class RemoveCorrelatedColumns:
    """
    Transform the data by removing correlated columns
    """
    def __init__(self, threshold=0.99):
        """
        Constructor of the object

        Args:
            threshold (float): the correlation absolute value threshold
        """
        self.threshold = threshold
        self.remove_mask = None

    def fit(self, X):
        """
        Fit the transformer.

        Args:
            X (np.array): the vectors

        Returns:
            RemoveCorrelatedColumns: the fitted object
        """
        corr = np.abs(np.corrcoef(X.T))
        corr[np.tril_indices(len(corr), k=0)] = np.nan
        self.remove_mask = ~np.any(corr > self.threshold, axis=1)
        return self

    def transform(self, X):
        """
        Transform data.

        Args:
            X (np.array): the vectors to transform

        Returns:
            X (np.array): the transformed data
        """
        return X[:, self.remove_mask]

def discrete_variable_mask(X, threshold=5):
    """
    Determine the mask of discrete variables.

    Args:
        X (np.array): the data
        threshold (int): threshold on unique attribute elements

    Returns:
        np.array: the binary mask of discrete variable columns
    """
    return np.apply_along_axis(lambda x: len(np.unique(x)), 0, X) <= threshold

def estimate_mutual_information(X,
                                y,
                                normalize=True,
                                n_repeats= 10,
                                mi_params=None):
    """
    Estimate the mutual information between variables and a target.

    Args:
        X (np.array): explanatory variables
        y (np.array): the target variables
        normalize (bool): whether to normalize the results
        n_repeats (int): the number of repetitions
        mi_params (dict): parameters of the MI estimation

    Returns:
        np.array: the array of MI scores for each column of X
    """
    if mi_params is None:
        mi_params = {'n_neighbors': 3}

    discrete_mask= discrete_variable_mask(X, threshold=5)

    mutuali = [mutual_info_regression(X, y,
                                **mi_params,
                                discrete_features=discrete_mask,
                                random_state=j) for j in range(n_repeats)]

    mutuali = np.mean(np.array(mutuali), axis=0)

    if normalize:
        mutuali = mutuali / np.mean(mutuali)

    return mutuali

def n_neighbors_func(X_base,
                        X_neighbors=None,
                        n_neighbors=5,
                        metric_tensor=None,
                        return_distance=False):
    """
    Determines the n closest neighbors with a metric tensor.

    Args:
        X_base (np.array): the base data
        X_neighbors (np.array): the neighbors data
        n_neighbors (int): the number of neighbors
        metric_tensor (np.array): the tensor to be used
        return_distance (bool): whether to return the distance

    Returns:
        np.array(, np.array): the indices of closest neighbors and optionally
                                their distances
    """
    if metric_tensor is None:
        metric_tensor = np.eye(X_base.shape[1])

    X_neighbors= X_neighbors if X_neighbors is not None else X_base

    X_diff= (X_base[:,None] - X_neighbors)

    distm= np.sqrt(np.einsum('ijk,ijk -> ij',
                            X_diff,
                            np.dot(X_diff, metric_tensor)).T).T

    results_ind= np.apply_along_axis(np.argsort,
                                        axis=1,
                                        arr=distm)[:,:(n_neighbors)]

    if not return_distance:
        return results_ind

    distances = distm[np.arange(distm.shape[0])[:,None], results_ind]
    return distances, results_ind


def psd_mean(matrices):
    """
    Estimate the mean of positive semi-definite matrices

    Args:
        matrices (iterable): the matrices

    Returns:
        np.array: the estimated mean matrix
    """
    result= np.mean(matrices, axis=0)

    result= fix_pd_matrix(result)

    return result

class MetricTensor:
    """
    Class representing a metric tensor.
    """

    def __init__(self,
                 metric_learning_method=None,
                 **_kwargs):
        """
        MetricTensor constructor

        Args:
            metric_learning_method (str): metric learning algorithm

        """
        self.metric_learning_method = metric_learning_method
        self.metric_tensor = None

    def get_params(self):
        """
        Get the paraemters of the object.

        Returns:
            dict: the parameters of the object
        """
        return {'metric_learning_method': self.metric_learning_method}

    def _train_metric_learning(self, X, y, method='NCA', *,
                                random_state=None, prior='identity'):
        """
        Do the metric learning training

        Args:
            X (np.array): the explanataory variables
            y (np.array): the target variable
            method (str): metric learning method
            random_state (int/None): random state to pass

        Returns:
            np.array: the metric tensor
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if method == 'ITML':
                return ITML_Supervised(random_state=random_state,
                                        prior=prior).fit(X, y)\
                                        .get_mahalanobis_matrix()
            if method == 'LSML':
                return LSML_Supervised().fit(X, y)\
                                        .get_mahalanobis_matrix()

            return NCA().fit(X, y)\
                                .get_mahalanobis_matrix()

    def tensor(self, X, y):
        """
        Create metric tensor for the parameters.

        Args:
            X (np.array): the explanataory variables
            y (np.array): the target variable

        Returns:
            np.array: the metric tensor
        """
        X_mod, index = np.unique(X, axis=0, return_index=True)
        y_mod = y[index]

        _logger.info("%s: executing metric learning with %s",
                        self.__class__.__name__,
                        self.metric_learning_method)

        if self.metric_learning_method == 'ITML':
            self.metric_tensor = self._train_metric_learning(X_mod,
                                                y_mod,
                                                self.metric_learning_method)
        elif self.metric_learning_method == 'rf':
            dissim= ClassifierImpliedDissimilarityMatrix().fit(X, y)\
                                            .dissimilarity_matrix(X)
            self.metric_tensor, _ = construct_tensor(X, dissim)
        elif self.metric_learning_method == 'LSML':
            self.metric_tensor = self._train_metric_learning(X_mod,
                                                y_mod,
                                                self.metric_learning_method)
        elif self.metric_learning_method == 'cov':
            self.metric_tensor = np.linalg.inv(fix_pd_matrix(np.cov(X.T)))
        elif self.metric_learning_method == 'cov_min':
            cov= np.cov(X[y == 1].T)
            self.metric_tensor = np.linalg.inv(fix_pd_matrix(cov))
        elif self.metric_learning_method == 'MI_weighted':
            mutuali= estimate_mutual_information(X, y)
            self.metric_tensor= np.diag(mutuali)
        elif self.metric_learning_method == 'id':
            self.metric_tensor= np.eye(len(X_mod[0]))
        elif self.metric_learning_method == 'ITML_mi':
            self.metric_tensor = self._train_metric_learning(X_mod,
                                                y_mod,
                                                self.metric_learning_method)
            mutuali= estimate_mutual_information(X, y)
            self.metric_tensor= np.matmul(self.metric_tensor, np.diag(mutuali))
        elif self.metric_learning_method == 'NCA':
            self.metric_tensor= NCA().fit(X_mod, y_mod)\
                                        .get_mahalanobis_matrix()
        elif self.metric_learning_method == 'gmean':
            matrices = [self._train_metric_learning(X_mod,
                                                y_mod,
                                                self.metric_learning_method,
                                                prior='random') for i in range(2)]

            self.metric_tensor= psd_mean(matrices)

        return self.metric_tensor

class MetricLearningMixin:
    """
    Mixin class for oversampling methods supporting custom metrics
    """

    def metric_tensor_from_nn_params(self, nn_params, X, y):
        """
        Determine the metric tensor from the given parameters.

        Args:
            nn_params (dict): the nearest neighbors parameters
            X (np.array): the explanatory variables
            y (np.array): the target variables

        Returns:
            np.array/None: the metric tensor
        """
        if nn_params.get('metric', None) == 'precomputed' \
                and nn_params.get('metric_tensor', None) is None:
            return MetricTensor(**nn_params).tensor(X, y)

        if nn_params.get('metric_tensor', None) is not None:
            return nn_params['metric_tensor']

        return None

    def get_params(self):
        """
        Get the parameters of the object.

        Returns:
            dict: the parameter dict
        """
        return {}

class NearestNeighborsWithMetricTensor:
    """
    NearestNeighbors driven by a metric tensor
    """
    def __init__(self,
                 n_neighbors=5,
                 *,
                 radius=1.0,
                 algorithm='auto',
                 leaf_size=30,
                 metric='minkowski',
                 p=2, # pylint: disable=invalid-name
                 metric_params=None,
                 metric_tensor=None,
                 n_jobs=None,
                 **_kwargs):
        """
        Constructor of the class

        Args:
            n_neighbors (int): the number of neighbors
            radius (float): the radius of neighbors
            algorithm (str): the algorithm to be used
            leaf_size (int): the size of leafs for quadtree estimation
            metric (str): the metric to be used
            p (int): the p-norm p value
            metric_params (dict): the parameters for metric learning
            metric_tensor (np.array): the metric tensor to be used
            n_jobs (int): the number of jobs
        """
        self.n_neighbors = n_neighbors
        self.radius = radius
        self.metric = metric
        self.metric_tensor = metric_tensor
        self.X_fitted = None

        self.nearestn= NearestNeighbors(n_neighbors=n_neighbors,
                                        radius=radius,
                                        algorithm=algorithm,
                                        leaf_size=leaf_size,
                                        metric=metric,
                                        p=p,
                                        metric_params=metric_params,
                                        n_jobs=n_jobs)

    def get_params(self):
        """
        Get the parameters of the object.

        Returns:
            dict: the parameters of the object
        """
        return {**self.nearestn.get_params(),
                "metric_tensor": self.metric_tensor}

    def fit(self, X):
        """
        Fit the nearest neighbors

        Args:
            X (np.array): the vectors to fit

        Returns:
            NearestNeighborsWithMetricTensor: the fitted object
        """
        _logger.info("%s: NN fitting with metric %s",
                        self.__class__.__name__,
                        self.metric)
        if self.metric != 'precomputed' or self.metric_tensor is None:
            self.nearestn.fit(X)
        else:
            self.X_fitted= X

        return self

    def kneighbors(self,
                    X=None,
                    n_neighbors=None,
                    return_distance=True):
        """
        Determine the k nearest neighbors

        Args:
            X (np.array): the vectors to determine the neighbors for
            n_neighbors (int): the number of neighbors
            return_distance (bool): whether to return the distance

        Returns:
            np.array(, np.array): the indices of nearest neighbors and
                                    optionally the distances
        """
        _logger.info("%s: kneighbors query %s",
                        self.__class__.__name__,
                        self.metric)

        if self.metric != 'precomputed' or self.metric_tensor is None:
            return self.nearestn.kneighbors(X, n_neighbors, return_distance)

        n_neighbors = coalesce(n_neighbors, self.n_neighbors)
        return n_neighbors_func(X, self.X_fitted,
                            metric_tensor=self.metric_tensor,
                            n_neighbors=n_neighbors,
                            return_distance=return_distance)

    def radius_neighbors(self,
                            X=None,
                            radius=None,
                            return_distance=True,
                            sort_results=False):
        """
        Determine the neighbors within a given radius.

        Args:
            X (np.array): the vectors to query the neighbors of
            radius (float/None): the radius
            return_distance (bool): whether to return the distances
            sort_results (bool): whether to sort the results
        """
        _logger.info("%s: radius neighbors query %s",
                        self.__class__.__name__,
                        self.metric)

        if self.metric != 'precomputed' or self.metric_tensor is None:
            return self.nearestn.radius_neighbors(X,
                                                    radius,
                                                    return_distance,
                                                    sort_results)

        diffs = (self.X_fitted[:,None] - X)
        distm = np.einsum('ijk,ijk -> ij',
                            diffs,
                            np.dot(diffs, self.metric_tensor)).T
        distm = np.sqrt(distm)

        results_dist= []
        results_ind= []

        for _, row in enumerate(distm):
            mask= np.where(row <= radius)[0]
            results_dist.append(row[mask])
            results_ind.append(mask)

        results_dist = np.array(results_dist, dtype=object)
        results_ind = np.array(results_ind, dtype=object)

        if return_distance:
            return results_dist, results_ind

        return results_ind



"""
Testing the Supervised_SMOTE.
"""

import numpy as np

from smote_variants import Supervised_SMOTE


from smote_variants.datasets import (load_all_min_noise,
                                     load_3_min_some_maj,
                                     load_illustration_2_class)

def test_specific():
    """
    Oversampler specific testing
    """
    obj = Supervised_SMOTE(th_lower=0.9, th_upper=0.95)

    dataset = load_illustration_2_class()

    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    dataset = load_3_min_some_maj()

    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    dataset = load_all_min_noise()

    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9], [9, 10], [10, 11],
                    [11, 12], [12, 13], [13, 14], [14, 15],
                    [15, 16], [16, 17]])
    y = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0



"""
All oversampling techniques.
"""

from ._nosmote import *
from ._smote import *
from ._smote_tomeklinks import *
from ._smote_enn import *
from ._borderline_smote import *
from ._adasyn import *
from ._ahc import *
from ._lle_smote import *
from ._distance_smote import *
from ._smmo import *
from ._polynom_fit_smote import *
from ._stefanowski import *
from ._safe_level_smote import *
from ._msmote import *
from ._de_oversampling import *
from ._smobd import *
from ._sundo import *
from ._msyn import *
from ._svm_balance import *
from ._trim_smote import *
from ._smote_rsb import *
from ._prowsyn import *
from ._sl_graph_smote import *
from ._nrsboundary_smote import *
from ._lvq_smote import *
from ._soi_cj import *
from ._rose import *
from ._smote_out import *
from ._smote_cosine import *
from ._selected_smote import *
from ._ln_smote import *
from ._mwmote import *
from ._pdfos import *
from ._ipade_id import *
from ._rwo_sampling import *
from ._neater import *
from ._deago import *
from ._gazzah import *
from ._mct import *
from ._adg import *
from ._smote_ipf import *
from ._kerneladasyn import *
from ._mot2ld import *
from ._v_synth import *
from ._oups import *
from ._smote_d import *
from ._smote_pso import *
from ._cure_smote import *
from ._somo import *
from ._ce_smote import *
from ._isomap_hybrid import *
from ._edge_det_smote import *
from ._cbso import *
from ._dbsmote import *
from ._asmobd import *
from ._assembled_smote import *
from ._sdsmote import *
from ._dsmote import *
from ._g_smote import *
from ._nt_smote import *
from ._lee import *
from ._spy import *
from ._smote_psobat import *
from ._mdo import *
from ._random_smote import *
from ._ismote import *
from ._vis_rst import *
from ._gasmote import *
from ._a_suwo import *
from ._smote_frst_2t import *
from ._and_smote import *
from ._nras import *
from ._amsco import *
from ._sso import *
from ._dsrbf import *
from ._ndo_sampling import *
from ._gaussian_smote import *
from ._kmeans_smote import *
from ._supervised_smote import *
from ._sn_smote import *
from ._ccr import *
from ._ans import *
from ._cluster_smote import *
from ._e_smote import *
from ._adoms import *
from ._symprod import *



"""
This module implements the ProWSyn method.
"""

import numpy as np

from scipy.linalg import circulant

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['ProWSyn']

class ProWSyn(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{prowsyn,
                        author="Barua, Sukarna
                        and Islam, Md. Monirul
                        and Murase, Kazuyuki",
                        editor="Pei, Jian
                        and Tseng, Vincent S.
                        and Cao, Longbing
                        and Motoda, Hiroshi
                        and Xu, Guandong",
                        title="ProWSyn: Proximity Weighted Synthetic
                                        Oversampling Technique for
                                        Imbalanced Data Set Learning",
                        booktitle="Advances in Knowledge Discovery
                                    and Data Mining",
                        year="2013",
                        publisher="Springer Berlin Heidelberg",
                        address="Berlin, Heidelberg",
                        pages="317--328",
                        isbn="978-3-642-37456-2"
                        }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 L=5,
                 theta=1.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            L (int): number of levels
            theta (float): smoothing factor in weight formula
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(L, "L", 1)
        self.check_greater_or_equal(theta, "theta", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = ss_params
        self.params = {'L': L,
                        'theta': theta}
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'L': [3, 5, 7],
                                  'theta': [0.1, 1.0, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples_in_clusters(self,
                                        X,
                                        Ps, # pylint: disable=invalid-name
                                        weights,
                                        n_to_sample):
        """
        Generate samples within the clusters.

        Args:
            X (np.array): all training vectors
            Ps (np.array): proximity clusters
            weights (np.array): weights of the clusters
            n_to_sample (int): the number of samples to generate
        """
        clusters_selected = self.random_state.choice(len(Ps),
                                                        n_to_sample,
                                                        p=weights)
        cluster_unique, cluster_count = np.unique(clusters_selected,
                                                    return_counts=True)
        #n_dim_original = self.n_dim
        samples = []
        for idx, cluster in enumerate(cluster_unique):
            cluster_vectors = X[Ps[cluster]]
            #self.n_dim = np.min([self.n_dim, cluster_vectors.shape[0]])
            samples.append(self.sample_simplex(X=cluster_vectors,
                                                indices=circulant(np.arange(len(cluster_vectors))),
                                                n_to_sample = cluster_count[idx]))
            #self.n_dim = n_dim_original

        return np.vstack(samples)

        # do the sampling, from each cluster proportionally to the distribution
        #samples = []
        #while len(samples) < n_to_sample:
        #    cluster_idx = self.random_state.choice(
        #        np.arange(len(weights)), p=weights)
        #    if len(Ps[cluster_idx]) > 1:
        #        random_idx1, random_idx2 = self.random_state.choice(
        #            Ps[cluster_idx], 2, replace=False)
        #        #samples.append(self.sample_between_points(
        #        #    X[random_idx1], X[random_idx2]))
        #        samples.append(X[random_idx1] + self.random_state.random_sample()
        #       *(X[random_idx2] - X[random_idx1]))
        #
        #return np.vstack(samples)

    def determine_clusters_proximity_levels(self, X, y):
        """
        Determine the clusters and proximity levels.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels

        Returns:
            list, np.array: the clusters and the proximity levels
        """
        X_maj = X[y == self.maj_label]
        P = np.where(y == self.min_label)[0] # pylint: disable=invalid-name
        Ps = [] # pylint: disable=invalid-name
        proximity_levels = []

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        # Step 3
        for idx in range(self.params['L']):
            if len(P) == 0:
                break
            # Step 3 a
            n_neighbors = np.min([len(P), self.n_neighbors])

            nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
            nnmt.fit(X[P])
            indices = nnmt.kneighbors(X_maj, return_distance=False)

            # Step 3 b
            P_i = np.unique(indices.flatten()) # pylint: disable=invalid-name

            # Step 3 c - proximity levels are encoded in the Ps list index
            Ps.append(P[P_i])
            proximity_levels.append(idx + 1)

            # Step 3 d
            P = np.delete(P, P_i) # pylint: disable=invalid-name

        # Step 4 and 5
        if len(P) > 0:
            Ps.append(P)
            proximity_levels.append(idx + 1)

        proximity_levels = np.array(proximity_levels)

        return Ps, proximity_levels

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and
                                    target labels
        """
        # Step 1 - a bit generalized
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        # Step 2
        Ps, proximity_levels = self.determine_clusters_proximity_levels(X, y) # pylint: disable=invalid-name

        # Step 6
        weights = np.exp(-self.params['theta']*(proximity_levels-1))

        # I think this can never happen
        #if not np.any(len(Ps_i) > 1 for Ps_i in Ps):
        #    return self.return_copies(X, y,
        #            "No clusters with at leats 2 samples")

        weights[np.array([idx for idx, Ps_i in enumerate(Ps)
                              if len(Ps_i) == 0], dtype=int)] = 0.0

        # weights is the probability distribution of sampling in the
        # clusters identified
        weights = weights/np.sum(weights)

        samples = self.generate_samples_in_clusters(X, Ps, weights,
                                                    n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'L': self.params['L'],
                'theta': self.params['theta'],
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the Oversampling base class.
"""

import time
import warnings

from ..config import suppress_internal_warnings

from ._base import (StatisticsMixin, ParametersMixin, RandomStateMixin,
                    coalesce_dict)
from ._metrictensor import MetricLearningMixin
from ._simplexsampling import SimplexSamplingMixin

from .._logger import logger
_logger = logger

__all__= ['OverSampling',
            'OverSamplingSimplex',
            'OverSamplingBase',
            'RandomSamplingMixin']

class RandomSamplingMixin(RandomStateMixin):
    """
    This is the random sampling mixin class
    """
    def __init__(self, random_state=None):
        """
        Constructor of the mixin.

        Args:
            random_state (int/None/np.random.RandomState): initial random
                                                            state
        """
        RandomStateMixin.__init__(self, random_state)

    def get_params(self, deep=False):
        """
        Return the parameter dictionary.

        Args:
            deep (bool): whether it should be a deep query

        Returns:
            dict: the parameter dictionary
        """
        return RandomStateMixin.get_params(self, deep)

    def sample_between_points(self, x_vector, y_vector):
        """
        Sample randomly along the line between two points.
        Args:
            x_vector (np.array): point 1
            y_vector (np.array): point 2
        Returns:
            np.array: the new sample
        """
        return x_vector + (y_vector - x_vector)\
                             * self.random_state.random_sample()

    def sample_between_points_componentwise(self, x_vector, y_vector, mask=None):
        """
        Sample each dimension separately between the two points.
        Args:
            x_vector (np.array): point 1
            y_vector (np.array): point 2
            mask (np.array): array of 0,1s - specifies which dimensions
                                to sample
        Returns:
            np.array: the new sample being generated
        """
        if mask is None:
            return x_vector + (y_vector - x_vector)*self.random_state.random_sample()

        return x_vector + (y_vector - x_vector)*self.random_state.random_sample()*mask

    def sample_by_jittering(self, x_vector, std):
        """
        Sample by jittering.
        Args:
            x_vector (np.array): base point
            std (float): standard deviation
        Returns:
            np.array: the new sample
        """
        return x_vector + (self.random_state.random_sample() - 0.5)*2.0*std

    def sample_by_jittering_componentwise(self, x_vector, std):
        """
        Sample by jittering componentwise.
        Args:
            x_vector (np.array): base point
            std (np.array): standard deviation
        Returns:
            np.array: the new sample
        """
        return x_vector + (self.random_state.random_sample(len(x_vector))-0.5)*2.0 * std

    def sample_by_gaussian_jittering(self, x_vector, std):
        """
        Sample by Gaussian jittering
        Args:
            x_vector (np.array): base point
            std (np.array): standard deviation
        Returns:
            np.array: the new sample
        """
        return self.random_state.normal(x_vector, std)

class OverSamplingBase(StatisticsMixin,
                   ParametersMixin,
                   MetricLearningMixin):
    """
    Base class of oversampling methods
    """

    categories = []

    cat_noise_removal = 'NR'
    cat_dim_reduction = 'DR'
    cat_uses_classifier = 'Clas'
    cat_sample_componentwise = 'SCmp'
    cat_sample_ordinary = 'SO'
    cat_sample_copy = 'SCpy'
    cat_memetic = 'M'
    cat_density_estimation = 'DE'
    cat_density_based = 'DB'
    cat_extensive = 'Ex'
    cat_changes_majority = 'CM'
    cat_uses_clustering = 'Clus'
    cat_borderline = 'BL'
    cat_application = 'A'
    cat_metric_learning = 'CD'

    def __init__(self, checks=None):
        """
        Constructor of the base class.

        Args:
            checks (dict): the check list
        """
        StatisticsMixin.__init__(self)
        ParametersMixin.__init__(self)
        MetricLearningMixin.__init__(self)
        checks_default = {'min_n_min': 2,
                          'check_np': True}
        self.checks = coalesce_dict(checks, checks_default)

    def det_n_to_sample(self, strategy, n_maj=None, n_min=None):
        """
        Determines the number of samples to generate
        Args:
            strategy (str/float): if float, the fraction of the difference
                                    of the minority and majority numbers to
                                    generate, like 0.1 means that 10% of the
                                    difference will be generated if str,
                                    like 'min2maj', the minority class will
                                    be upsampled to match the cardinality
                                    of the majority class
            n_maj (int/None): the number of majority samples
            n_min (int/None): the number of minority samples

        Returns:
            int: the number of samples to generate
        """
        if n_maj is None:
            n_maj = self.class_stats[self.maj_label]
        if n_min is None:
            n_min = self.class_stats[self.min_label]

        if isinstance(strategy, (int, float)):
            return max([0, int((n_maj - n_min)*strategy)])

        raise ValueError(f"{self.__class__.__name__} Value {strategy} "\
                            "for parameter strategy is not supported")

    def fit_resample(self, X, y):
        """
        Alias of the function "sample" for compatibility with imbalanced-learn
        pipelines

        Args:
            X (np.array): the feature vectors
            y (np.array): the target labels

        Returns:
            np.array, np.array: the oversampled dataset
        """
        return self.sample(X, y)

    def sampling_algorithm(self, X, y):
        """
        The algorithm to be implemented.

        Args:
            X (np.array): features
            y (np.array): labels

        Returns:
            np.array, np.array: the oversampled dataset
        """
        return X, y

    def sample(self, X, y):
        """
        Sampling interface function.

        Args:
            X (np.array): features
            y (np.array): labels

        Returns:
            np.array, np.array: the oversampled dataset
        """
        _logger.info("%s: Running sampling via %s",
                        self.__class__.__name__,
                        self.descriptor())

        self.class_label_statistics(y)

        for key, item in self.checks.items():
            if key == 'min_n_min':
                if self.class_stats[self.min_label] <= item:
                    msg = f"{self.__class__.__name__}: Too few minority samples"\
                            " for sampling"
                    _logger.info(msg)
                    return X.copy(), y.copy()
            if key == 'min_n_dim':
                if X.shape[1] < item:
                    _logger.info("%s: not enough dimensions %d",
                                self.__class__.__name__, X.shape[1])
                    return X.copy(), y.copy()

        return self.sampling_algorithm(X, y)

    def return_copies(self, X, y, msg):
        """
        Returns copies of the data with logger message.

        Args:
            X (np.array): features
            y (np.array): labels
            msg (str): the logging message

        Returns:
            np.array, np.array: the oversampled dataset
        """
        _logger.info("%s: returning copies for %s",
                        self.__class__.__name__, msg)

        if not suppress_internal_warnings():
            warnings.warn(f"{self.__class__.__name__}: returning copies for {msg}")

        return X.copy(), y.copy()

    def sample_with_timing(self, X, y):
        """
        Execute the sampling with timing.

        Args:
            X (np.array): features
            y (np.array): labels
            msg (str): the logging message

        Returns:
            np.array, np.array: the oversampled dataset
        """
        begin = time.time()

        X_samp, y_samp = self.sample(X, y)

        _logger.info("%s: runtime: %f", self.__class__.__name__,
                                                (time.time() - begin))
        return X_samp, y_samp

    def preprocessing_transform(self, X):
        """
        Transforms new data according to the possible transformation
        implemented by the function "sample".

        Args:
            X (np.array): features

        Returns:
            np.array: transformed features
        """
        return X

    def get_params(self, deep=False):
        """
        Returns the parameters of the object as a dictionary.

        Returns:
            dict: the parameters of the object
        """
        _ = deep
        return {}

    def set_params(self, **params):
        """
        Set parameters

        Args:
            params (dict): dictionary of parameters
        """

        for key, value in params.items():
            setattr(self, key, value)

        return self

    def descriptor(self):
        """
        The descriptor of the class

        Returns:
            str: JSON description of the current sampling object
        """
        return str((self.__class__.__name__, str(self.get_params())))

    def __str__(self):
        """
        The string representation

        Returns:
            str: the descriptor
        """
        return self.descriptor()


class OverSampling(OverSamplingBase, RandomSamplingMixin):
    """
    The oversampling base class.
    """
    def __init__(self, random_state=None, checks=None):
        OverSamplingBase.__init__(self, checks)
        RandomSamplingMixin.__init__(self, random_state)

    def get_params(self, deep=False):
        return {**RandomSamplingMixin.get_params(self, deep),
                'class_name': self.__class__.__name__}


class OverSamplingSimplex(OverSamplingBase, SimplexSamplingMixin):
    """
    The oversampling simplex base class.
    """
    def __init__(self,
                *,
                n_dim=2,
                simplex_sampling='uniform',
                within_simplex_sampling='random',
                gaussian_component=None,
                random_state=None,
                checks=None):
        OverSamplingBase.__init__(self, checks)

        if checks is not None and 'simplex_dim' in checks:
            if n_dim != checks['simplex_dim']:
                warnings.warn(f"Simplex dimensions {n_dim} not supported "\
                                f"with the method {self.__class__.__name__} "\
                                f"forcing n_dim={checks['simplex_dim']}")
                n_dim = checks['simplex_dim']

        SimplexSamplingMixin.__init__(self,
                                        n_dim=n_dim,
                                        simplex_sampling=simplex_sampling,
                                        within_simplex_sampling=within_simplex_sampling,
                                        gaussian_component=gaussian_component,
                                        random_state=random_state)

    def get_params(self, deep=False):
        return {**SimplexSamplingMixin.get_params(self, deep),
                'class_name': self.__class__.__name__}



"""
Tests oversamplers with a few samples.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_some_min_some_maj

dataset = load_some_min_some_maj()

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_some_min_some_maj(smote_class):
    """
    Tests an oversampler with only a few samples.

    Args:
        smote_class (class): the oversampler class
    """
    X, y = smote_class().sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
Testing the oversamplers and noise removals under normal conditions.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_normal
from .additional_objs import additional_objs

dataset = load_normal()

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_normal(smote_class):
    """
    Tests an oversmampler

    Args:
        smote_class (class): an oversampler class
    """
    X, y = smote_class().sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", additional_objs)
def test_normal_additional(smote_obj):
    """
    Tests some oversamplers with additional parameterizations.

    Args:
        smote_obj (obj): an oversampling object
    """
    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("nf_class", sv.get_all_noisefilters())
def test_normal_noisefilters(nf_class):
    """
    Tests a noise remove technique.

    Args:
        nf_class (class): a noise removal class
    """
    X, y = nf_class().remove_noise(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
This module implements some imbalanced classification metrics.
"""

import numpy as np

from sklearn.metrics import roc_auc_score, log_loss

__all__ = ['prediction_labels',
            'calculate_atoms',
            'calculate_label_scores',
            'calculate_prob_scores',
            'calculate_all_scores',
            'all_scores']

all_scores = ['acc', 'sens', 'spec', 'ppv', 'npv', 'fpr', 'fdr',
                'fnr', 'bacc', 'gacc', 'f1', 'mcc', 'l', 'ltp', 'lfp', 'lfn',
                'ltn', 'lp', 'ln', 'uc', 'informedness', 'markedness', 'p_top20',
                'brier', 'log_loss', 'auc']

def prediction_labels(probabilities_maj):
    """
    Determine the labels from the probabilities.

    Args:
        probabilities_maj (np.array): the majority probabilities

    Returns:
        np.array: the labels row-by-row
    """
    labels = (probabilities_maj > 0.5) * 1
    equals = probabilities_maj == 0.5
    indices = np.where(equals)[0]
    if len(indices) <= 1:
        return labels

    half = int(len(indices)/2)
    labels[indices[:half]] = 0
    labels[indices[half:]] = 1
    return labels

def calculate_atoms(test_labels, predicted_labels, min_label=1):
    """
    Calculate the atoms used for the measures.

    Args:
        test_labels (np.array): the true labels
        predicted_labels (np.array): the predicted labels
        min_label (int): the minority label

    Returns:
        dict: the atoms
    """

    atoms = {}

    equals = np.equal(test_labels, predicted_labels)
    not_equals = np.logical_not(equals)

    min_sample = test_labels == min_label
    maj_sample = np.logical_not(min_sample)

    atoms['tp'] = int(np.sum(np.logical_and(equals, min_sample)))
    atoms['tn'] = int(np.sum(np.logical_and(equals, maj_sample)))

    atoms['fp'] = int(np.sum(np.logical_and(not_equals, maj_sample)))
    atoms['fn'] = int(np.sum(np.logical_and(not_equals, min_sample)))

    return atoms

def _log_score(multiplier, value):
    """
    Calculates a log score and returs None if not computable.

    Args:
        multiplier (float): the multiplier
        value (float): the value to take the log of

    Returns:
        float: the score
    """
    if value > 0:
        log_value = np.log(value)
    else:
        log_value = np.nan
    if not np.isfinite(log_value):
        return None
    return float(multiplier * log_value)

def _log_score_div(numerator, denominator):
    """
    Calculates a log score and returs None if not computable.

    Args:
        nominator (float): the nominator
        denominator (float): the denominator

    Returns:
        float: the score
    """
    if denominator > 0:
        return _log_score(numerator, numerator / denominator)
    return None

def calculate_label_scores(atoms):
    """
    Calculate scores from labels.

    Args:
        atoms (dict): the atomic scores

    Returns:
        dict: the label scores
    """
    atoms['p'] = atoms['tp'] + atoms['fn']
    atoms['n'] = atoms['fp'] + atoms['tn']

    atoms['acc'] = (atoms['tp'] + atoms['tn']) / (atoms['p'] + atoms['n'])
    atoms['sens'] = atoms['tp'] / atoms['p']
    atoms['spec'] = atoms['tn'] / atoms['n']
    if atoms['tp'] + atoms['fp'] > 0:
        atoms['ppv'] = atoms['tp'] / (atoms['tp'] + atoms['fp'])
    else:
        atoms['ppv'] = 0.0
    if atoms['tn'] + atoms['fn'] > 0:
        atoms['npv'] = atoms['tn'] / (atoms['tn'] + atoms['fn'])
    else:
        atoms['npv'] = 0.0
    atoms['fpr'] = 1.0 - atoms['spec']
    atoms['fdr'] = 1.0 - atoms['ppv']
    atoms['fnr'] = 1.0 - atoms['sens']
    atoms['bacc'] = (atoms['sens'] + atoms['spec'])/2.0
    atoms['gacc'] = float(np.sqrt(atoms['sens']*atoms['spec']))
    atoms['f1'] = 2 * atoms['tp'] / (2 * atoms['tp'] + atoms['fp'] + atoms['fn'])

    tp_fp = (atoms['tp'] + atoms['fp'])
    tp_fn = (atoms['tp'] + atoms['fn'])
    tn_fp = (atoms['fp'] + atoms['tn'])
    tn_fn = (atoms['fn'] + atoms['tn'])

    mcc_num = atoms['tp']*atoms['tn'] - atoms['fp']*atoms['fn']
    mcc_denom = float(np.prod([tp_fp, tp_fn, tn_fp, tn_fn]))

    if mcc_denom == 0:
        atoms['mcc'] = None
    else:
        atoms['mcc'] = float(mcc_num/np.sqrt(mcc_denom))

    atoms['l'] = float((atoms['p'] + atoms['n']) * np.log(atoms['p'] + atoms['n']))

    atoms['ltp'] = _log_score_div(atoms['tp'], tp_fp * tp_fn)
    atoms['lfp'] = _log_score_div(atoms['fp'], tp_fp * tn_fp)
    atoms['lfn'] = _log_score_div(atoms['fn'], tp_fn * tn_fn)
    atoms['ltn'] = _log_score_div(atoms['tn'], tn_fp * tn_fn)

    atoms['lp'] = float(atoms['p'] * np.log(atoms['p']/(atoms['p'] + atoms['n'])))
    atoms['ln'] = float(atoms['n'] * np.log(atoms['n']/(atoms['p'] + atoms['n'])))

    items = [atoms['ltp'], atoms['lfp'], atoms['lfn'], atoms['ltn']]

    if np.all([item is not None for item in items]):
        uc_num = atoms['l'] + np.sum(items)
        uc_denom = atoms['l'] + atoms['lp'] + atoms['ln']
        atoms['uc'] = uc_num / uc_denom
    else:
        atoms['uc'] = None

    atoms['informedness'] = atoms['sens'] + atoms['spec'] - 1.0
    atoms['markedness'] = atoms['ppv'] + atoms['npv'] - 1.0

    return atoms

def calculate_prob_scores(test_labels, probabilities, min_label=1):
    """
    Calculate scores from probabilities.

    Args:
        test_labels (np.array): the true labels
        probabilities (np.array): the probabilities
        min_label (int): the minority label

    Returns:
        dict: the probability scores
    """
    results = {}

    thres = max(int(0.2*len(test_labels)), 1)
    results['p_top20'] = float(np.sum(test_labels[:thres] == min_label)/thres)
    results['brier'] = float(np.mean((probabilities - test_labels)**2))
    results['log_loss'] = float(log_loss(test_labels, probabilities))
    results['auc'] = float(roc_auc_score(test_labels, probabilities))

    return results

def calculate_all_scores(test_labels, probabilities, min_label=1):
    """
    Calculate all scores.

    Args:
        test_labels (np.array): the true labels
        probabilities (np.array): the probabilities
        min_label (int): the minority label

    Returns:
        dict: all scores
    """
    pred_labels = prediction_labels(probabilities)
    results = calculate_atoms(test_labels, pred_labels, min_label)
    results = calculate_label_scores(results)
    results = {**results, **calculate_prob_scores(test_labels,
                                                  probabilities,
                                                  min_label)}
    return results



"""
Tests the configurations
"""

from smote_variants.config import (suppress_external_warnings,
                                   suppress_internal_warnings,
                                   DEFAULT_SUPPRESS_EXTERNAL_WARNINGS,
                                   DEFAULT_SUPPRESS_INTERNAL_WARNINGS)

def test_external_warnings():
    """
    Tests the external warnings settings
    """

    suppress_external_warnings(True)

    assert suppress_external_warnings()

    suppress_external_warnings(False)

    assert not suppress_external_warnings()

    suppress_external_warnings(DEFAULT_SUPPRESS_EXTERNAL_WARNINGS)

def test_internal_warnings():
    """
    Tests the internal warnings settings
    """

    suppress_internal_warnings(True)

    assert suppress_internal_warnings()

    suppress_internal_warnings(False)

    assert not suppress_internal_warnings()

    suppress_internal_warnings(DEFAULT_SUPPRESS_INTERNAL_WARNINGS)



"""
This module implements the SMOTE_OUT method.
"""

import numpy as np

from ..base import coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['SMOTE_OUT']

class SMOTE_OUT(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_out_smote_cosine_selected_smote,
                      title={SMOTE-Out, SMOTE-Cosine, and Selected-SMOTE: An
                                enhancement strategy to handle imbalance in
                                data level},
                      author={Fajri Koto},
                      journal={2014 International Conference on Advanced
                                Computer Science and Information System},
                      year={2014},
                      pages={280-284}
                    }
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): parameter of the NearestNeighbors component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self, *, X, X_maj, X_min, minority_indices,
                            n_to_sample, maj_indices, min_indices):
        """
        Generate samples

        Args:
            X (np.array): all training vectors
            X_maj (np.array): majority vectors
            X_min (np.array): minority vectors
            minority_indices (np.array): the minority indices
            n_to_sample (int): number of samples to generate
            maj_indices (np.array): majority neighborhood structure
            min_indices (np.array): minority neighborhood structure

        Returns:
            np.array: the generated samples
        """
        base_ind = self.random_state.choice(np.arange(len(minority_indices)),
                                            n_to_sample)

        u = X[minority_indices[base_ind]] # pylint: disable=invalid-name
        neigh_ind = self.random_state.choice(np.arange(0, maj_indices.shape[1]),
                                             n_to_sample)
        v = X_maj[maj_indices[base_ind, neigh_ind]] # pylint: disable=invalid-name
        uu = u + self.random_state.random_sample(u.shape) * 0.3 * (u - v) # pylint: disable=invalid-name
        min_neigh_ind = self.random_state.choice(np.arange(1, min_indices.shape[1]),
                                                 n_to_sample)
        x = X_min[min_indices[base_ind, min_neigh_ind]] # pylint: disable=invalid-name
        return x + self.random_state.random_sample(x.shape) * 0.5 * (uu - x)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])
        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        minority_indices = np.where(y == self.min_label)[0]

        # nearest neighbors among minority points
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        n_neighbors_min = min([len(X_min), self.n_neighbors+1])
        nn_min= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors_min,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
        nn_min.fit(X_min)

        min_indices = nn_min.kneighbors(X_min, return_distance=False)
        # nearest neighbors among majority points
        n_neighbors_maj = min([len(X_maj), self.n_neighbors+1])
        nn_maj= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors_maj,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
        nn_maj.fit(X_maj)
        maj_indices = nn_maj.kneighbors(X_min, return_distance=False)

        samples = self.generate_samples(X=X, X_maj=X_maj, X_min=X_min,
                    minority_indices=minority_indices, n_to_sample=n_to_sample,
                    maj_indices=maj_indices, min_indices=min_indices)


        # generate samples
        #samples = []
        #for _ in range(n_to_sample):
        #    # implementation of Algorithm 1 in the paper
        #    random_idx = self.random_state.choice(np.arange(len(minority_indices)))
        #    u = X[minority_indices[random_idx]]
        #    v = X_maj[self.random_state.choice(maj_indices[random_idx])]
        #    dif1 = u - v
        #    uu = u + self.random_state.random_sample()*0.3*dif1
        #    x = X_min[self.random_state.choice(min_indices[random_idx][1:])]
        #    dif2 = uu - x
        #    w = x + self.random_state.random_sample()*0.5*dif2
        #
        #    samples.append(w)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the KernelADASYN method.
"""

import numpy as np

from sklearn.decomposition import PCA

from ..base import fix_density, cov
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['KernelADASYN']

class KernelADASYN(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{kernel_adasyn,
                            author={Tang, B. and He, H.},
                            booktitle={2015 IEEE Congress on Evolutionary
                                        Computation (CEC)},
                            title={KernelADASYN: Kernel based adaptive
                                    synthetic data generation for
                                    imbalanced learning},
                            year={2015},
                            volume={},
                            number={},
                            pages={664-671},
                            keywords={learning (artificial intelligence);
                                        pattern classification;
                                        sampling methods;KernelADASYN;
                                        kernel based adaptive synthetic
                                        data generation;imbalanced
                                        learning;standard classification
                                        algorithms;data distribution;
                                        minority class decision rule;
                                        expensive minority class data
                                        misclassification;kernel based
                                        adaptive synthetic over-sampling
                                        approach;imbalanced data
                                        classification problems;kernel
                                        density estimation methods;Kernel;
                                        Estimation;Accuracy;Measurement;
                                        Standards;Training data;Sampling
                                        methods;Imbalanced learning;
                                        adaptive over-sampling;kernel
                                        density estimation;pattern
                                        recognition;medical and
                                        healthcare data learning},
                            doi={10.1109/CEC.2015.7256954},
                            ISSN={1089-778X},
                            month={May}}

    Notes:
        * The method of sampling was not specified, Markov Chain Monte Carlo
            has been implemented.
        * Not prepared for improperly conditioned covariance matrix.
    """

    categories = [OverSampling.cat_density_estimation,
                  OverSampling.cat_extensive,
                  OverSampling.cat_borderline,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 k=5,
                 *,
                 nn_params={},
                 h=1.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            k (int): number of neighbors in the nearest neighbors component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            h (float): kernel bandwidth
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(k, 'k', 1)
        self.check_greater(h, 'h', 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.k = k # pylint: disable=invalid-name
        self.nn_params = nn_params
        self.h = h # pylint: disable=invalid-name
        self.n_jobs = n_jobs

        self.mcmc_params = {'burn_in' : 1000,
                            'periods' : 50}

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'k': [5, 7, 9],
                                  'h': [0.01, 0.02, 0.05, 0.1, 0.2,
                                        0.5, 1.0, 2.0, 10.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def p_x(self, x_vec, X_min, r_score):
        """
        Returns minority density value at x

        Args:
            x_vec (np.array): feature vector

        Returns:
            float: density value
        """
        result = 1.0/(len(X_min)*self.h)
        result = result*(1.0/(np.sqrt(2*np.pi)*self.h)**len(X_min[0]))

        exp_term = np.exp(-0.5*np.linalg.norm(x_vec - X_min, axis=1)**2/self.h)
        return result*np.inner(r_score, exp_term)

    def mcmc(self, X_min, r_score, covariance, n_to_sample):
        """
        Do the Markov Chain Monte Carlo sampling.

        Args:
            X_min (np.array): the minority samples
            r_score (np.array): the majority scores
            covariance (np.array): the covariance matrix
            n_to_sample (int): the number of samples to generate
        """
        samples = []
        iteration = 0

        # parameters of the Monte Carlo sampling


        # starting Markov-Chain Monte Carlo for sampling
        x_old = X_min[self.random_state.choice(np.where(r_score > 0)[0])]
        p_old = self.p_x(x_old, X_min, r_score)

        # Cholesky decomposition
        L_mat = np.linalg.cholesky(covariance) # pylint: disable=invalid-name

        while len(samples) < n_to_sample:
            x_new = x_old + \
                np.dot(self.random_state.normal(size=len(x_old)), L_mat)
            p_new = self.p_x(x_new, X_min, r_score)

            alpha = p_new/p_old
            if self.random_state.random_sample() < alpha:
                x_old = x_new
                p_old = p_new

            iteration = iteration + 1
            if (iteration % self.mcmc_params['periods'] == 0
                                and iteration > self.mcmc_params['burn_in']):
                samples.append(x_old)

        return np.vstack(samples)

    def iterated_sampling(self, *,
                        covariance,
                        X, y,
                        X_min,
                        r_score,
                        n_to_sample):
        """
        Try to carry out the sampling and reduce the dimensions if it
        does not succeed.

        Args:
            covariance (np.array): the covariance matrix
            X (np.array): the feature vectors
            y (np.array): the target labels
            X_min (np.array): the minority samples
            r_score (np.array): the majority score
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array, np.array: the oversampled dataset
        """
        if len(covariance) > 1 and np.linalg.cond(covariance) > 10000:
            _logger.info("%s: reducing dimensions due to improperly "\
                            "conditioned covariance matrix",
                        self.__class__.__name__)

            if X.shape[1] <= 2:
                return self.return_copies(X, y, "matrix is ill conditioned")

            n_components = int(np.rint(len(covariance)/2))

            pca = PCA(n_components=n_components)
            X_trans = pca.fit_transform(X) # pylint: disable=invalid-name

            kernela = KernelADASYN(proportion=self.proportion,
                              k=self.k,
                              nn_params=self.nn_params,
                              h=self.h,
                              random_state=self._random_state_init)

            X_samp, y_samp = kernela.sample(X_trans, y)

            return pca.inverse_transform(X_samp), y_samp

        samples = self.mcmc(X_min, r_score, covariance, n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting the nearest neighbors model
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=min([len(X_min), self.k+1]),
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        indices = nnmt.kneighbors(X_min, return_distance=False)

        # computing majority score
        r_score = np.array([np.sum(y[indices[i][1:]] == self.maj_label)
                      for i in range(len(X_min))])

        if np.sum(r_score > 0) < 2:
            return self.return_copies(X, y, "majority score is 0 for most "\
                                                "vectors")

        r_score = fix_density(r_score)

        # covariance is used to generate a random sample in the neighborhood
        covariance = cov(X_min[r_score > 0], rowvar=False)

        return self.iterated_sampling(covariance=covariance,
                                        X=X, y=y,
                                        X_min=X_min,
                                        r_score=r_score,
                                        n_to_sample=n_to_sample)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'k': self.k,
                'nn_params': self.nn_params,
                'h': self.h,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the MSMOTE method.
"""

import numpy as np

from ..base import coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['MSMOTE']

class MSMOTE(OverSampling):
    """
    References:
        * BibTex::

            @inproceedings{msmote,
                             author = {Hu, Shengguo and Liang,
                                 Yanfeng and Ma, Lintao and He, Ying},
                             title = {MSMOTE: Improving Classification
                                        Performance When Training Data
                                        is Imbalanced},
                             booktitle = {Proceedings of the 2009 Second
                                            International Workshop on
                                            Computer Science and Engineering
                                            - Volume 02},
                             series = {IWCSE '09},
                             year = {2009},
                             isbn = {978-0-7695-3881-5},
                             pages = {13--17},
                             numpages = {5},
                             url = {https://doi.org/10.1109/WCSE.2009.756},
                             doi = {10.1109/WCSE.2009.756},
                             acmid = {1682710},
                             publisher = {IEEE Computer Society},
                             address = {Washington, DC, USA},
                             keywords = {imbalanced data, over-sampling,
                                        SMOTE, AdaBoost, samples groups,
                                        SMOTEBoost},
                            }

    Notes:
        * The original method was not prepared for the case when all
            minority samples are noise.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_borderline,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): control parameter of the nearest neighbor
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self, *, X_min, X, indices, sample_type, n_to_sample):
        """
        Generate samples.

        Args:
            X_min (np.array): the minority samples
            X (np.array): all training vectors
            indices (np.array): the neighborhood structure
            sample_type (np.array): the sample types
            n_to_sample (int): number of samples to generate

        Returns:
            np.array: the generated samples
        """
        not_noise_mask = sample_type != 'NOI'
        X_not_noise = X_min[not_noise_mask] # pylint: disable=invalid-name
        st_not_noise = sample_type[not_noise_mask]
        ind_not_noise = indices[not_noise_mask]

        base_indices = self.random_state.choice(X_not_noise.shape[0],
                                                n_to_sample)

        neighbor_indices = np.array([1] * n_to_sample)
        neighbor_indices[st_not_noise[base_indices] == 'SEC'] = \
            self.random_state.choice(np.arange(1, indices.shape[1]),
                                     np.sum(st_not_noise[base_indices] == 'SEC'))

        neighbor_indices = ind_not_noise[base_indices,
                                         neighbor_indices]

        base_vectors = X_not_noise[base_indices]
        neighbor_vectors = X[neighbor_indices]

        samples = base_vectors + (neighbor_vectors - base_vectors) \
                       * self.random_state.random_sample(base_vectors.shape)

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # fitting the nearest neighbors model
        n_neighbors = min([len(X), self.n_neighbors+1])

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        indices = nnmt.kneighbors(X_min, return_distance=False)

        n_p = np.sum(y[indices[:, 1:]] == self.min_label, axis=1)
        sample_type = np.array(['BOR'] * n_p.shape[0])
        sample_type[n_p == (n_neighbors - 1)] = 'SEC'
        sample_type[n_p == 0] = 'NOI'

        if np.all(sample_type == 'NOI'):
            return self.return_copies(X, y, "All samples are noise")

        samples = self.generate_samples(X_min=X_min,
                                        X=X,
                                        indices=indices,
                                        sample_type=sample_type,
                                        n_to_sample=n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the LVQ_SMOTE method.
"""
import warnings

import numpy as np

from sklearn.cluster import KMeans

from ..config import suppress_external_warnings
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['LVQ_SMOTE']

class LVQ_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @inproceedings{lvq_smote,
                              title={LVQ-SMOTE – Learning Vector Quantization
                                    based Synthetic Minority Over–sampling
                                    Technique for biomedical data},
                              author={Munehiro Nakamura and Yusuke Kajiwara
                                     and Atsushi Otsuka and Haruhiko Kimura},
                              booktitle={BioData Mining},
                              year={2013}
                            }

    Notes:
        * This implementation is only a rough approximation of the method
            described in the paper. The main problem is that the paper uses
            many datasets to find similar patterns in the codebooks and
            replicate patterns appearing in other datasets to the imbalanced
            datasets based on their relative position compared to the codebook
            elements. What we do is clustering the minority class to extract
            a codebook as kmeans cluster means, then, find pairs of codebook
            elements which have the most similar relative position to a
            randomly selected pair of codebook elements, and translate nearby
            minority samples from the neighborhood one pair of codebook
            elements to the neighborood of another pair of codebook elements.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_application]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params={},
                 n_clusters=10,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_clusters (int): number of clusters in vector quantization
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(n_clusters, "n_clusters", 3)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.n_clusters = n_clusters
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_clusters': [4, 8, 12]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_samples(self, n_to_sample, codebook, X_min, indices):
        """
        Generate samples.

        Args:
            n_to_sample (int): the number of samples to generate
            codebook (np.array): the codebook
            X_min (np.array): the minority samples
            indices (np.array): the neighborhood structure

        Returns:
            np.array: the generated samples
        """

        # do the sampling
        samples = []
        while len(samples) < n_to_sample:
            # randomly selecting a pair of codebook elements
            codes = self.random_state.choice(list(range(len(codebook))),
                                                    2,
                                                    replace=False)
            diff = codebook[codes[0]] - codebook[codes[1]]
            optimum = (np.inf, None)
            # finding another pair of codebook elements with similar offset
            for idx, cb_i in enumerate(codebook):
                for jdx, cb_j in enumerate(codebook):
                    if (codes[0] not in [idx, jdx]
                                    and codes[1] not in [idx, jdx]):
                        ddiff = np.linalg.norm(diff - (cb_i - cb_j))
                        if ddiff < optimum[0]:
                            optimum = (ddiff, self.random_state.choice([idx, jdx]))

            # translating a random neighbor of codebook element min_0 to
            # the neighborhood of point_0
            idx = self.random_state.randint(len(indices[optimum[1]]))

            sample = X_min[indices[optimum[1]][idx]]

            samples.append(codebook[codes[0]] + \
                                        (sample - codebook[optimum[1]]))

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # clustering X_min to extract codebook
        n_clusters = min([len(X_min), self.n_clusters])
        kmeans = KMeans(n_clusters=n_clusters,
                        random_state=self._random_state_init)
        with warnings.catch_warnings():
            if suppress_external_warnings():
                warnings.simplefilter("ignore")
            kmeans.fit(X_min)
        codebook = kmeans.cluster_centers_

        # get nearest neighbors of minority samples to codebook samples
        n_neighbors = min([len(X_min), self.n_neighbors])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        indices = nnmt.kneighbors(codebook, return_distance=False)

        samples = self.generate_samples(n_to_sample,
                                        codebook,
                                        X_min,
                                        indices)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_clusters': self.n_clusters,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Tests oversamplers with a few samples.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import (load_1_dim,
                                     load_illustration_2_class,
                                     load_normal,
                                     load_same_num,
                                     load_some_min_some_maj,
                                     load_1_min_some_maj,
                                     load_2_min_some_maj,
                                     load_3_min_some_maj,
                                     load_4_min_some_maj,
                                     load_5_min_some_maj,
                                     load_1_min_1_maj,
                                     load_repeated,
                                     load_all_min_noise,
                                     load_separable,
                                     load_linearly_dependent,
                                     load_alternating,
                                     load_high_dim)

oversamplers = [sv.SMOTE()]

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_1_dim(smote_obj):
    """
    Testing oversamplers with 1 minority sample.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_1_dim()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_1_min_some_maj(smote_obj):
    """
    Testing oversamplers with 1 minority sample.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_1_min_some_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_1_min_1_maj(smote_obj):
    """
    Testing oversamplers with 1 minority and 1 majority sample.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_1_min_1_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_repeated(smote_obj):
    """
    Testing oversamplers with repeated samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_repeated()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_2_min_some_maj(smote_obj):
    """
    Testing oversamplers with 2 minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_2_min_some_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_3_min_some_maj(smote_obj):
    """
    Testing oversamplers with 3 minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_3_min_some_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_4_min_some_maj(smote_obj):
    """
    Testing oversamplers with 4 minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_4_min_some_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_5_min_some_maj(smote_obj):
    """
    Testing oversamplers with 5 minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_5_min_some_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_all_min_noise(smote_obj):
    """
    Testing oversamplers with all minority samples being noise.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_all_min_noise()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_alternating(smote_obj):
    """
    Testing oversamplers with alternating minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_alternating()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_high_dim(smote_obj):
    """
    Testing an oversampler with high dimensionality data.

    Args:
        smote_obj (obj): an oversampler obj
    """
    dataset = load_high_dim()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])
    assert len(X) > 0
    assert X.shape[1] == smote_obj\
                                .preprocessing_transform(dataset['data']).shape[1]
    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_illustration(smote_obj):
    """
    Testing oversamplers with illustration data.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_illustration_2_class()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_linearly_dependent(smote_obj):
    """
    Testing oversamplers with 2 minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_linearly_dependent()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_normal(smote_obj):
    """
    Tests an oversmampler

    Args:
        smote_obj (obj): an oversampler obj
    """
    dataset = load_normal()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_reproducibility(smote_obj):
    """
    Tests the reproducibility of oversampling.

    Args:
        smote_obj (obj): an oversampling obj
    """
    dataset = load_normal()

    X_normal = dataset['data'] # pylint: disable=invalid-name
    y_normal = dataset['target']

    X_orig = X_normal.copy()
    y_orig = y_normal.copy()

    X_a, y_a = smote_obj.__class__(random_state=5).sample(X_normal, y_normal)
    oversampler = smote_obj.__class__(random_state=5)
    X_b, y_b = oversampler.sample(X_normal, y_normal)
    X_c, y_c = smote_obj.__class__(**oversampler.get_params()).sample(X_normal, y_normal)

    assert np.array_equal(X_a, X_b)
    assert np.array_equal(X_b, X_c)
    assert np.array_equal(X_orig, X_normal)

    assert np.array_equal(y_a, y_b)
    assert np.array_equal(y_b, y_c)
    assert np.array_equal(y_orig, y_normal)

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_same_num(smote_obj):
    """
    Tests oversamplers with equalized data.

    Args:
        smote_obj (obj): an oversampling obj
    """
    dataset = load_same_num()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_separable(smote_obj):
    """
    Testing oversamplers with 3 minority samples.

    Args:
        smote_obj (obj): an oversampler obj.
    """
    dataset = load_separable()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_some_min_some_maj(smote_obj):
    """
    Tests an oversampler with only a few samples.

    Args:
        smote_obj (obj): the oversampler obj
    """
    dataset = load_some_min_some_maj()

    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", oversamplers)
def test_parameters(smote_obj):
    """
    Test the parameterization.

    Args:
        smote_obj (obj): an oversampling object
    """
    random_state = np.random.RandomState(5)

    par_comb = smote_obj.__class__.parameter_combinations()

    original_parameters = random_state.choice(par_comb)
    oversampler = smote_obj.__class__(**original_parameters)
    parameters = oversampler.get_params()

    assert all(v == parameters[k] for k, v in original_parameters.items())



"""
This module implements the CBSO method.
"""

import numpy as np
from scipy.linalg import circulant

from ..base import coalesce_dict, fix_density, coalesce
from ..base import (NearestNeighborsWithMetricTensor,
                             pairwise_distances_mahalanobis)
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['CBSO']

class CBSO(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{cbso,
                            author="Barua, Sukarna
                            and Islam, Md. Monirul
                            and Murase, Kazuyuki",
                            editor="Lu, Bao-Liang
                            and Zhang, Liqing
                            and Kwok, James",
                            title="A Novel Synthetic Minority Oversampling
                                    Technique for Imbalanced Data Set
                                    Learning",
                            booktitle="Neural Information Processing",
                            year="2011",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="735--744",
                            isbn="978-3-642-24958-7"
                            }

    Notes:
        * Clusters containing 1 element induce cloning of samples.
    """

    categories = [OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_density_based,
                  OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 C_p=1.3,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): the simplex sampling parameters
            C_p (float): used to set the threshold of clustering
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater(C_p, "C_p", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.C_p = C_p # pylint: disable=invalid-name
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'C_p': [0.8, 1.0, 1.3, 1.6]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def do_clustering(self, X_min, nn_params):
        """
        Do the clustering of minority samples.

        Args:
            X_min (np.array): minority samples
            nn_params (dict): nearest neighbor parameters

        Returns:
            np.array, np.array: clusters, labels
        """
        # do the clustering
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=2,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        d_avg = np.mean(nnmt.kneighbors(X_min)[0][:, 1])
        T_h = d_avg * self.C_p # pylint: disable=invalid-name

        # initiating clustering
        clusters = [np.array([i]) for i in range(len(X_min))]
        distm = pairwise_distances_mahalanobis(X_min,
                                            tensor=nn_params.get('metric_tensor', None))

        # setting the diagonal of the distance matrix to infinity
        for idx in range(len(distm)):
            distm[idx, idx] = np.inf

        # starting the clustering iteration
        while True:
            # finding the cluster pair with the smallest distance
            min_coord = np.where(distm == np.min(distm))
            merge_a = min_coord[0][0]
            merge_b = min_coord[1][0]

            # check termination conditions
            if distm[merge_a, merge_b] > T_h or len(distm) == 1:
                break

            # merging the clusters
            clusters[merge_a] = np.hstack(
                [clusters[merge_a], clusters[merge_b]])
            # removing one of them
            del clusters[merge_b]
            # adjusting the distances in the distance matrix
            distm[merge_a] = np.min(np.vstack([distm[merge_a], distm[merge_b]]), axis=0)
            distm[:, merge_a] = distm[merge_a]
            # removing the row and column corresponding to one of the
            # merged clusters
            distm = np.delete(distm, merge_b, axis=0)
            distm = np.delete(distm, merge_b, axis=1)
            # updating the diagonal
            for idx in range(len(distm)):
                distm[idx, idx] = np.inf

        # extracting cluster labels
        labels = np.zeros(len(X_min)).astype(int)
        for idx, cluster in enumerate(clusters):
            for jdx in cluster:
                labels[jdx] = idx

        return clusters, labels

    def generate_samples_in_clusters(self,
                                    *,
                                    clusters,
                                    X_min,
                                    n_to_sample,
                                    weights):
        """
        Generate samples within the clusters.

        Args:
            clusters (np.array): the clusters
            X_min (np.array): the minority samples
            n_to_sample (int): number of samples to generate
            weights (np.array): the minority point density

        Returns:
            np.array: the generated samples
        """

        cluster_weights = np.array([np.sum(weights[cluster]) for cluster in clusters])
        cluster_weights = cluster_weights / np.sum(cluster_weights)

        cluster_indices = self.random_state.choice(np.arange(len(clusters)),
                                                    n_to_sample,
                                                    p=cluster_weights)

        cluster_unique, cluster_count = np.unique(cluster_indices,
                                                    return_counts=True)

        samples = []
        for idx, cluster in enumerate(cluster_unique):
            cluster_vectors = X_min[clusters[cluster]]
            within_cluster_weights = weights[clusters[cluster]]
            within_cluster_weights = fix_density(within_cluster_weights)

            #if len(cluster_vectors) >= self.n_dim:
            samples.append(self.sample_simplex(X=cluster_vectors,
                                        indices=circulant(np.arange(cluster_vectors.shape[0])),
                                        n_to_sample=cluster_count[idx],
                                        base_weights=within_cluster_weights))
            #else:
            #    sample_indices = self.random_state.choice(np.arange(len(cluster_vectors)),
            #                                                cluster_count[idx],
            #                                                p=within_cluster_weights)
            #    samples.append(cluster_vectors[sample_indices])

        # original implementation
        # do the sampling
        #samples = []
        #while len(samples) < n_to_sample:
        #    idx = self.random_state.choice(np.arange(len(X_min)), p=weights)
        #    if len(clusters[labels[idx]]) <= 1:
        #        samples.append(X_min[idx])
        #        continue
        #
        #    random_idx = self.random_state.choice(clusters[labels[idx]])
        #    while random_idx == idx:
        #        random_idx = self.random_state.choice(
        #            clusters[labels[idx]])
        #    samples.append(self.sample_between_points(
        #        X_min[idx], X_min[random_idx]))

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params,
                                                                        X, y)

        # fitting nearest neighbors model to find neighbors of minority points
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=self.n_neighbors + 1,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        # extracting the number of majority neighbors
        weights = [np.sum(y[ind[i][1:]] == self.maj_label)
                   for i in range(len(X_min))]

        weights = fix_density(weights)

        clusters, _ = self.do_clustering(X_min, nn_params)

        samples = self.generate_samples_in_clusters(clusters=clusters,
                                                    n_to_sample=n_to_sample,
                                                    X_min=X_min,
                                                    weights=weights)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'C_p': self.C_p,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}
